import {
  getContextClient,
  setContextClient
} from "./chunk-K6BWBPDV.mjs";

// src/index.ts
export * from "@trpc/client";

// src/createTRPCSvelte.ts
import {
  createInfiniteQuery,
  createMutation,
  createQuery
} from "@tanstack/svelte-query";
import {
  createTRPCUntypedClient
} from "@trpc/client";
import {
  createFlatProxy,
  createRecursiveProxy
} from "@trpc/server/shared";
import { BROWSER as BROWSER2 } from "esm-env";
import { writable } from "svelte/store";

// src/internals/getArrayQueryKey.ts
function getArrayQueryKey(path, input, type) {
  const hasInput = typeof input !== "undefined";
  const hasType = type && type !== "any";
  if (!hasInput && !hasType)
    return path.length ? [path] : [];
  const inputAndType = {};
  if (hasInput)
    inputAndType.input = input;
  if (hasType)
    inputAndType.type = type;
  return [path, inputAndType];
}

// src/shared/utils.ts
import {
  useQueryClient
} from "@tanstack/svelte-query";
import { BROWSER } from "esm-env";

// src/utils/splitUserOptions.ts
function splitUserOptions(options) {
  if (options === void 0) {
    return [];
  }
  const trpcOptions = {
    context: options.context,
    signal: options.signal
  };
  delete options.context;
  delete options.signal;
  return [trpcOptions, options];
}

// src/shared/utils.ts
var queryTypes = {
  invalidate: "any",
  prefetch: "query",
  prefetchInfinite: "infinite",
  fetch: "query",
  fetchInfinite: "infinite",
  refetch: "any",
  cancel: "any",
  reset: "any",
  setData: "query",
  setInfiniteData: "infinite",
  getData: "query",
  getInfiniteData: "infinite"
};
function callUtilMethod(trpc, client, path, method, args) {
  if (!BROWSER) {
    client = useQueryClient();
  }
  const queryType = queryTypes[method];
  const queryKey = getArrayQueryKey(path, args[0], queryType);
  switch (method) {
    case "prefetch":
    case "fetch":
    case "prefetchInfinite":
    case "fetchInfinite": {
      const joinedPath = path.join(".");
      const options = args[1];
      const [trpcOptions, tanstackQueryOptions] = splitUserOptions(options);
      const queryFn = queryType === "query" ? () => trpc.query(joinedPath, args[0], trpcOptions) : (context) => {
        const input = { ...args[0], cursor: context.pageParam };
        return trpc.query(joinedPath, input, trpcOptions);
      };
      return client[method + "Query"]({
        ...tanstackQueryOptions,
        queryKey,
        queryFn
      });
    }
    case "invalidate":
      return client.invalidateQueries(queryKey, args[1], args[2]);
    case "refetch":
      return client.refetchQueries(
        getArrayQueryKey(path, args[0], "query"),
        args[1],
        args[2]
      );
    case "cancel":
      return client.cancelQueries(queryKey, args[1], args[2]);
    case "reset":
      return client.resetQueries(queryKey, args[1], args[2]);
    case "setData":
    case "setInfiniteData":
      return client.setQueryData(queryKey, args[1], args[2]);
    case "getData":
    case "getInfiniteData":
      return client.getQueryData(queryKey, args[1]);
    default:
      throw new TypeError(`trpc.${path}.${method} is not a function`);
  }
}

// src/createTRPCSvelte.ts
var clientMethods = {
  query: [1, "query"],
  mutation: [0, "any"],
  infiniteQuery: [1, "infinite"],
  prefetchQuery: [1, "query"],
  prefetchInfiniteQuery: [1, "infinite"],
  subscription: [1, "query"]
};
function createHooksInternalProxy(client, opts) {
  const queryClient = opts.queryClient;
  return createFlatProxy((firstPath) => {
    switch (firstPath) {
      case "queryClient": {
        return queryClient;
      }
    }
    return createRecursiveProxy(({ path, args: unknownArgs }) => {
      path.unshift(firstPath);
      const method = path.pop();
      const joinedPath = path.join(".");
      const methodData = clientMethods[method];
      if (!methodData) {
        const utils = path.pop();
        if (utils === "utils") {
          return callUtilMethod(
            client,
            queryClient,
            path,
            method,
            unknownArgs
          );
        }
        throw new TypeError(`trpc.${joinedPath}.${method} is not a function`);
      }
      const args = unknownArgs;
      const [optionIndex, queryType] = methodData;
      const options = args[optionIndex];
      const [trpcOptions, tanstackQueryOptions] = splitUserOptions(options);
      const key = getArrayQueryKey(
        path,
        method === "mutation" ? void 0 : args[0],
        queryType
      );
      const enabled = tanstackQueryOptions?.enabled !== false;
      switch (method) {
        case "query":
          return createQuery({
            ...tanstackQueryOptions,
            enabled,
            queryKey: key,
            queryFn: () => client.query(joinedPath, args[0], trpcOptions)
          });
        case "mutation": {
          return createMutation({
            ...tanstackQueryOptions,
            mutationKey: key,
            mutationFn: (variables) => client.mutation(joinedPath, variables, trpcOptions)
          });
        }
        case "subscription": {
          const state = writable("idle");
          const data = writable(void 0);
          const error = writable(
            void 0
          );
          const enabled2 = (options?.enabled ?? true) && BROWSER2;
          if (!enabled2) {
            return {
              state,
              data,
              error,
              unsubscribe: () => void 0
            };
          }
          const subscription = client.subscription(joinedPath, args[0], {
            onStarted: () => {
              state.set("running");
            },
            onData: (data_) => {
              data.set(data_);
            },
            onError: (error_) => {
              error.set(error_);
            }
          });
          const unsubscribe = () => {
            state.set("stopped");
            subscription.unsubscribe();
          };
          return {
            state,
            data,
            error,
            unsubscribe
          };
        }
        case "infiniteQuery":
          return createInfiniteQuery({
            ...tanstackQueryOptions,
            enabled,
            queryKey: key,
            queryFn: (context) => {
              const input = { ...args[0], cursor: context.pageParam };
              return client.query(joinedPath, input, trpcOptions);
            }
          });
        case "prefetchQuery": {
          return queryClient.prefetchQuery({
            ...tanstackQueryOptions,
            queryKey: key,
            queryFn: () => client.query(joinedPath, args[0], trpcOptions)
          });
        }
        case "prefetchInfiniteQuery": {
          return queryClient.prefetchInfiniteQuery({
            ...tanstackQueryOptions,
            queryKey: key,
            queryFn: (context) => {
              const input = { ...args[0], cursor: context.pageParam };
              return client.query(joinedPath, input, trpcOptions);
            }
          });
        }
        default:
          throw new TypeError(`trpc.${joinedPath}.${method} is not a function`);
      }
    });
  });
}
function createTRPCSvelteClient(opts) {
  const client = createTRPCUntypedClient(opts);
  const proxy = createHooksInternalProxy(client, opts);
  return proxy;
}
function createTRPCSvelte() {
  const createClient = createTRPCSvelteClient;
  const setContextClient2 = setContextClient;
  const getContextClient2 = getContextClient;
  return {
    createClient,
    setContextClient: setContextClient2,
    getContextClient: getContextClient2
  };
}
export {
  createTRPCSvelte,
  createTRPCSvelteClient
};
